% 
% SVM
% 
% S. Canu, Sept 2014, scanu@insa-rouen.fr

% sample size
n = 20; % up to 200000 !
n = 100;

rand('seed',2);
Xi = 4*rand(n,2);
% add useless variables to see what...
q = 0; % up to 180;
%q = 150;

Xi = [Xi 4*rand(n,q)];
[n,p] = size(Xi);
% inspired from 
% http://home.comcast.net/~tom.fawcett/public_html/ML-gallery/pages/linear_
% Support%20vector%20machines.html
%bt = -1.74;  wt = [1.87 ; -1];
% the true line
bt = -6;
wt = [4 ; -1];
yi = sign(wt(1) * Xi(:,1) + wt(2) * Xi(:,2) + bt);

figure(1)
clf
set(gcf,'Color',[1,1,1])
aa = plot(Xi(:,1),Xi(:,2),'or');
hold on
bb = plot(Xi(find(yi==1),1),Xi(find(yi==1),2),'ob');
% how to draw a the true line
x1 = 0;
y1 = (-bt-(wt(1)*x1))/wt(2);
x2 = 4;
y2 = (-bt-(wt(1)*x2))/wt(2);
ch = plot([x1 x2],[y1 y2],'g','LineWidth',2);
axis([0 4 0  4]);


%% Max margin SVM

fprintf('   CVX Primal - Max margin SVM   : ')
tic
cvx_begin
   cvx_precision best
   cvx_quiet(true)
   variables v(p) a m 
   maximize(  m  )
   subject to
        yi.*(Xi*v + a) >= m;
        v'*v <= 1;
cvx_end
toc

% drawing the lines
if q==0
    x1 = 0;
    y1 = (-a-(v(1)*x1))/v(2);
    z1 = (m-a-(v(1)*x1))/v(2);
    zm1 = (-m-a-(v(1)*x1))/v(2);
    x2 = 4;
    y2 = (-a-(v(1)*x2))/v(2);
    z2 = (m-a-(v(1)*x2))/v(2);
    zm2 = (-m-a-(v(1)*x2))/v(2);
dd =     plot([x1 x2],[y1 y2],'k');
 dm=   plot([x1 x2],[z1 z2],':k');
    plot([x1 x2],[zm1 zm2],':k')
    axis([0 4 0  4]);
    
    set(gca,'FontSize',18,'FontName','Arial','XTick',[ ],'YTick',[ ],'Box','on');
    legend([aa bb ch dd dm],'class 1','class 2','true','SVM','Margin')
end

%% Classical L2 SVM

fprintf('   CVX Primal - Classical L2 SVM : ')
tic
cvx_begin
   cvx_precision best
   cvx_quiet(true)
   variables w(p) b 
   dual variables pi 
   minimize( .5*w'*w )
   subject to
      pi :  yi.*(Xi*w + b) >= 1;
cvx_end
toc

% 1) KKT check - 
% A active set == support vectors
A = find(yi.*(Xi*w + b) <= 1.00001);
cA = length(A);

DA = diag(yi(A));

KKT = [eye(p)        -Xi(A,:)'*DA     zeros(p,1)
          -DA*Xi(A,:)   zeros(cA)     -yi(A)
          zeros(1,p)    -yi(A)'            zeros(1)];
Kb = [zeros(p,1) ; -ones(cA,1) ;0];

sol = KKT\Kb;

% if you know A, you can find the solution at the cost 
% of solving a linear system
%
% check
if n==20
    fprintf('  Check CVX provides KKT (w, alpha, b) \n')
    disp([[w;pi(A);b] sol]);
end

% 2)  L2 SVM Dual

G = (yi*yi').*(Xi*Xi');
e = ones(n,1);

fprintf('   CVX Dual   - Classical L2 SVM : ')
tic
cvx_begin
   cvx_precision best
   cvx_quiet(true)
   variable a(n)  
   dual variables de dp
   minimize( 1/2*a'*G*a - e'*a )
   subject to
      de : yi'*a == 0;
      dp :  a >= 0;
cvx_end
toc


%% standart formulation
%  X = QUADPROG(H,f,A,b) attempts to solve the quadratic programming     problem:
% 
%            min 0.5*x'*H*x + f'*x   subject to:  A*x <= b 
%             x    

H = [eye(p)];
H(p+1,p+1) = 0;
f = zeros(p+1,1);
A = -[diag(yi)*Xi yi];
bb = -ones(n,1);

options = optimset('Display','none','TolFun',1e-8);
fprintf('quadpg Primal - Classical L2 SVM : ')
tic
x = quadprog(H,f,A,bb,[],[],[],[],[],options);
toc

l = eps^.5;
G = G + l*eye(n);   %  the secret to make it work

if n < 1001
    options.MaxIter = 2000; % default is 200
    fprintf('quadprog Dual - Classical L2 SVM : ')
    tic
    ad = quadprog(G,-e,[],[],yi',0,zeros(n,1),inf*ones(n,1),zeros(n,1),options);
    toc
end

%% Quadratic programing with mon QP

fprintf('   MonQP Dual - Classical L2 SVM : ')
tic
[alpha, bqp, pos] = monqp(G,e,yi,0,inf,l,0);
toc


%% check equivalences
if n==20
    fprintf(' Check that Primal = Dual (alpha)\n')
    aqp = zeros(n,1);
    aqp(pos) = alpha;
    wqp = Xi(pos,:)'*(yi(pos).*alpha);
    disp([a pi ad aqp])
    % 4) where is b ?
    fprintf(' Check that Primal = Dual (b,w)\n')
    disp([b de x(3)  bqp]);
    disp([w  Xi'*(yi.*a) x(1:2) wqp]);
end


%% adding noise and slack variables

yi(1) = -1;
yi(n) = 1;
C = 1000;

fprintf('\n   CVX Primal - L2 SVM + Error   : '),
tic
cvx_begin
   cvx_precision best
   cvx_quiet(true)
   variables w(p) b xi(n)
   dual variables pi px
   minimize( .5*w'*w  + C*sum(xi))
   subject to
      pi :  yi.*(Xi*w + b) >= 1 - xi;
      px: xi >= 0;
cvx_end
toc

%% Quadprog in the Primal
H = [eye(p)];
H(p+1+n,p+1+n) = 0;
f = [zeros(p+1,1);C*ones(n,1)];
A = -[diag(yi)*Xi yi eye(n)];
bb = -ones(n,1);

if n < 1001
    fprintf('quadpg Primal - L2 SVM + Error   : ')
    tic
    [xe,fval,exitflag,output,lambda] = quadprog(H,f,A,bb,[],[],[-inf*ones(p+1,1);zeros(n,1)],[],[],options);
    toc
end

%% KKT with slack variables

% 1) KKT check - 
% A active set == support vectors yi.*(Xi*w + b)) == 1 (never true in the computer world)
A = find(round(1000*yi.*(Xi*w + b)) == 1000);
cA = length(A);
% I_C error points with alpha = C and xi > 0
I_C =  find(yi.*(Xi*w + b) >= C-.00001); % 
I_C =  find(xi >= .00001);                      % xi  <neq>  0
cC = length(I_C);

Cxy = C*Xi(I_C,:)'*yi(I_C);

DA = diag(yi(A));

KKT = [eye(p)        -Xi(A,:)'*DA     zeros(p,1)
          -DA*Xi(A,:)   zeros(cA)     -yi(A)
          zeros(1,p)    -yi(A)'            zeros(1)];
Kb = [Cxy ; -ones(cA,1) ; C*sum(yi(I_C))];

sol = KKT\Kb;

%% SVM in the dual

G = (yi*yi').*(Xi*Xi');
e = ones(n,1);

fprintf('   CVX dual   - L2 SVM + Error   : '),
tic
cvx_begin
   cvx_precision best
   cvx_quiet(true)
   variable a(n)  
   dual variables de dp dC
   minimize( 1/2*a'*G*a - e'*a )
   subject to
      de : yi'*a == 0;
      dp :  a >= 0;
      dC :  a <= C;
cvx_end
toc

%
[b de];
[w  Xi'*(yi.*a)];

%%  dual in the standart form

if n < 1001
    fprintf('quadprog Dual - L2 SVM + Error   : ')
    tic
    adn = quadprog( G,-e,[],[],yi',0,zeros(n,1),C*ones(n,1),[],options);
    toc
end

%  mon QP
fprintf('   MonQP dual - L2 SVM + Error   : '),
tic
[alpha, b, pos] = monqp(G,e,yi,0,C,l,0);
toc


%% 


